#
#  JMClient.py
#  JunkMatcher
#
#  Created by Benjamin Han on 2/2/05.
#  Copyright (c) 2005 Benjamin Han. All rights reserved.
#

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#!/usr/bin/env python

from pluginConsts import *

import socket, fcntl, os, sys, time, logging

userName = os.environ['USER']


class JMClient:
    def __init__ (self, serverPortFN = None, port = None, block = True):
        """Connects to JMServer if it exists; set block to True to try POLL_SERVER_MAX times.
        self.port is NOT -1 if JMServer is not found."""
        if port is not None:
            # we're sure about the right port - nothing needs to be verified!
            self.port = port
            self.connected = False
            return
        
        elif serverPortFN is None:
            raise Exception('At least one of serverPortFN and port should not be None.')
        
        for i in range(POLL_SERVER_MAX):
            try:                
                s = open(serverPortFN).read()
                if s != 'ERR':
                    self.port = int(s)

                    # we use user's login name to distinguish if the server is this user's
                    # server (to avoid cross-talk)
                    if self.connect() and self.send(JM_CMD_PING, userName) == JM_REPLY_OK:
                        return
                else:
                    # even when block == True, return if JMServer reports 'ERR' in
                    # server.port file
                    self.port = -1
                    return                    
            except:
                pass

            if not block:
                self.port = -1
                return
                
            time.sleep(POLL_SERVER_INTVL)

        self.port = -1
    
    def connect (self):
        """Attempts to connect to JMServer at self.port"""
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect(('', self.port))
            self.connected = True
            return True
        except:
            self.socket.close()
            self.connected = False
            return False

    def send (self, cmd, arg = None):
        """Sends a command and its argument to JMServer, and waits for and returns
        the reply.

        Returns a string if a JMServer indeed responded; otherwise returns None."""
        if self.connected or self.connect():
            if arg:
                self.socket.sendall('%s%s%d%s%s' % (cmd, DELIMITER, len(arg), DELIMITER, arg))
            else:
                self.socket.sendall(cmd)
            
            reply = ''
            while True:
                reply += self.socket.recv(1024)
                if reply.endswith(END_MARKER): break

            self.socket.close()
            self.connected = False
            
            return reply.rstrip()
        
        else:
            return None


if __name__ == '__main__':

    if len(sys.argv) < 3:
        print '* Usage: ./JMClient.py <server port FN> <cmd> [arg]'
        sys.exit(1)
    
    jmc = JMClient(serverPortFN = sys.argv[1])
    if jmc.port != -1:
        if len(sys.argv) == 4:
            print jmc.send(sys.argv[2], sys.argv[3])
        else:
            print jmc.send(sys.argv[2])
    else:
        print 'Cannot find JMServer.'
