#
#  JMServer.py
#  JunkMatcher
#
#  Created by Benjamin Han on 2/2/05.
#  Copyright (c) 2005 Benjamin Han. All rights reserved.
#

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#!/usr/bin/env python

import sys, SocketServer, socket, threading

if len(sys.argv) != 2:
    sys.exit(1)

# sys.argv[1] is the Engine path!
sys.path.insert(0, sys.argv[1])

from consts import *
from utilities import *
from pluginConsts import *
from Matcher import *


# if a stale server.port file is found... (it might happen if Mail.app crashed and
# didn't tell JMServer to stop)
serverPortFN = TMP_PATH + 'server.port'
if os.path.exists(serverPortFN):
    # ... find the leftover server and tell it to stop
    jmc = __import__('JMClient').JMClient(serverPortFN = serverPortFN, block = False)
    if jmc.port != -1:
        jmc.send(JM_CMD_STOP_SYNC)

    # we asked - now try to nuke all JMServer processes by force
    pidSet = __import__('nukeJMServer').nukeJMServer()
    if pidSet:
        NSLog(u'JMServer.py processes killed: %s' % ', '.join(pidSet))

# initialize a matcher object, in parallel
matcher = None
matcherInitDoneEvent = threading.Event()
def initMatcher ():
    global matcher

    # remember to allocate AutoreleasePool!
    pool = NSAutoreleasePool.new()

    try:
        matcher = Matcher(True)
        matcher.recycleLogWhenItsDue()
        NSLog(u'Matcher initialization completed.')
    except Exception, e:
        matcher = None
        printException('Matcher initialization FAILED:', e)
        
    matcherInitDoneEvent.set()
    
threading.Thread(target = initMatcher).start()

# so we don't do certain things when we're in the middle of matching
matchingFinishedEvent = threading.Event()
matchingFinishedEvent.set()

# we halt matching when we're configuring the backend
serverOpFinishedEvent = threading.Event()
serverOpFinishedEvent.set()

# serverStateLock is to ensure the atomicity of calls to
# serverOpStarts() and matchingStarts()
serverStateLock = threading.Lock()

def serverOpStarts ():
    serverStateLock.acquire()
    serverOpFinishedEvent.clear()
    matchingFinishedEvent.wait()
    serverStateLock.release()

def matchingStarts ():
    serverStateLock.acquire()
    serverOpFinishedEvent.wait()
    matchingFinishedEvent.clear()
    serverStateLock.release()


class JMHandler (SocketServer.BaseRequestHandler):
    """Handler thread for JMServer"""
    def _send (self, reply):
        try:
            self.request.sendall(reply + END_MARKER)
        except:
            pass
        self.request.close()
    
    def handle (self):
        # --------------------------------------------------
        # get the entire input over the socket
        # --------------------------------------------------
        
        try:
            input = self.request.recv(MAX_CMD_LEN)
        except:
            return   # timeout or other errors: ignore this cmd

        arg = None
        try:
            pos1 = input.index(DELIMITER)
            cmd = input[:pos1]
            pos2 = input.index(DELIMITER, pos1 + 1)
            arg = input[pos2+1:]

            realLen = int(input[pos1 + 1 : pos2])
            l = len(input) - pos2 - 1

            while l < realLen:
                # we still have stuff not received
                try:
                    input = self.request.recv(1024)
                    arg += input
                    l += len(input)
                except:
                    pass
        except:
            cmd = input

        # --------------------------------------------------
        # execute the command
        # --------------------------------------------------

        # CAUTION: arg might still have END_MARKER

        if cmd == JM_CMD_PING:
            # ping the server
            if os.environ['USER'] == arg:
                self._send(JM_REPLY_OK)
            else:
                self._send(JM_REPLY_ERR)

        elif cmd == JM_CMD_MSG:
            # IMPORTANT: this is threaded - allocate NSAutoreleasePool!
            pool = NSAutoreleasePool.new()

            matcherInitDoneEvent.wait()
            if matcher is None:
                self._send(JM_REPLY_ERR)
                NSLog(u'Mathcer is AWOL')
            else:
                try:
                    # declare the server is matching (blocking configuration attempts)
                    matchingStarts()

                    msg = Message(arg)
                    
                    matchResult = matcher.run(msg)

                    if matchResult.verdict is True:
                        if globalObjects.prefs.junkMessage:
                            self._send(JM_REPLY_POS_JUNK)
                        else:
                            self._send(JM_REPLY_POS)                            

                    else:
                        # either the message is whitelisted or clean
                        self._send(JM_REPLY_NEG)
                        
                except Exception, e:
                    printException(u'Exception when JMServer is matching', e)

                matchingFinishedEvent.set()

        elif cmd == JM_CMD_STOP or cmd == JM_CMD_STOP_SYNC:
            if cmd == JM_CMD_STOP:
                self._send(JM_REPLY_OK)
            
            # IMPORTANT: this is threaded - allocate NSAutoreleasePool!
            pool = NSAutoreleasePool.new()            
            matchingFinishedEvent.wait()

            if matcher:
                NSLog(u'Matcher finalizing...')
                matcher.finalize()
            
            NSLog(u'JunkMatcher stopped.')

            try:
                os.remove(TMP_PATH + 'server.port')
            except:
                pass

            # for synchronized stop - reply only when we're about to leave
            if cmd == JM_CMD_STOP_SYNC:
                self._send(JM_REPLY_OK)
            
            os._exit(0)


        else:            
            # declare the server is configuring (blocking matching attempts)
            serverOpStarts()

            # IMPORTANT: this is threaded - allocate NSAutoreleasePool!
            pool = NSAutoreleasePool.new()

            try:
                if cmd == JM_CMD_RECYCLE:
                    
                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        self._send(JM_REPLY_OK)
                        matcher.recycleLog()
                        NSLog(u'Log recycled by JunkMatcher.')

                elif cmd == JM_CMD_SAFESITES:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.safeSitesPattern = loadSafeSitesPattern()

                    NSLog(u'JunkMatcher safe sites reloaded.')

                elif cmd == JM_CMD_WHITELIST:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.whitelist.load()

                    NSLog(u'JunkMatcher whitelist reloaded.')

                elif cmd == JM_CMD_RECIPIENTPATTERNS:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.recipientPatterns.load()

                    NSLog(u'JunkMatcher recipient patterns reloaded.')

                elif cmd == JM_CMD_SAFEIPS:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.safeIPs.load()

                    NSLog(u'JunkMatcher safe IPs reloaded.')

                elif cmd == JM_CMD_ABOOK:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.addressSet = __import__('addressBook').getAddressesFromABook()
                
                    NSLog(u'Address Book reloaded by JunkMatcher.')

                elif cmd == JM_CMD_HTMLTAGS:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.htmlTags.load()
                
                    NSLog(u'HTML tags reloaded by JunkMatcher.')

                elif cmd == JM_CMD_PREFS_RECYCLEDAYS:
                    globalObjects.prefs.recycleDays = int(arg)

                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        if matcher.recycleLogWhenItsDue():
                            self._send(JM_REPLY_RECYCLED)
                        else:
                            self._send(JM_REPLY_OK)
                
                        NSLog(u'Log recycle days reloaded by JunkMatcher.')

                elif cmd == JM_CMD_PREFS_SITEDBLIMIT:                    
                    siteLimit = int(arg)
                    globalObjects.prefs.siteLimit = siteLimit
                    if globalObjects.siteDB.setSizeLimit(siteLimit):
                        self._send(JM_REPLY_PRUNED)
                    else:
                        self._send(JM_REPLY_OK)
                
                    NSLog(u'SiteDB limit reloaded by JunkMatcher.')
                    
                elif cmd == JM_CMD_PREFS_MODE:
                    self._send(JM_REPLY_OK)

                    prefs = globalObjects.prefs
                    l = arg.split(' ')
                    prefs.mode = int(l[0])
                    prefs.modeArgs = l[1:]

                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        matcher.setMode(prefs.mode, *prefs.modeArgs)                
                        NSLog(u'Matching mode reloaded by JunkMatcher.')
                    
                elif cmd == JM_CMD_PREFS_JUNK:
                    self._send(JM_REPLY_OK)
            
                    globalObjects.prefs.junkMessage = (arg == '1')
                
                    NSLog(u'Junk flag reloaded by JunkMatcher.')
                    
                elif cmd == JM_CMD_LOAD_PROPERTIES:
                    self._send(JM_REPLY_OK)

                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        matcher.tests.properties.load()                
                        NSLog(u'JunkMatcher properties reloaded.')
                
                elif cmd == JM_CMD_LOAD_META:
                    self._send(JM_REPLY_OK)

                    globalObjects.metaPatterns.load()

                    NSLog(u'JunkMatcher meta patterns reloaded.')

                elif cmd == JM_CMD_LOAD_PATTERNS:
                    self._send(JM_REPLY_OK)

                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        matcher.tests.patterns.load()                
                        NSLog(u'JunkMatcher patterns reloaded.')

                elif cmd == JM_CMD_LOAD_TESTS:
                    self._send(JM_REPLY_OK)

                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        matcher.tests.load()
                        NSLog(u'JunkMatcher tests reloaded.')

                elif cmd == JM_CMD_GET_SITEDB:
                    self._send(str(globalObjects.siteDB))

                elif cmd == JM_CMD_SET_SITEDB:
                    self._send(JM_REPLY_OK)

                    if arg is not None: arg = arg
                    else: arg = ''   # this will clear out the entire SiteDB!

                    globalObjects.siteDB.loadFromString(arg)

                elif cmd == JM_CMD_LOAD_SITEDB:
                    self._send(JM_REPLY_OK)
            
                    # WARNING: this will overwrite the sites collected during the current runtime!
                    globalObjects.siteDB.load()
                
                    NSLog(u'JunkMatcher siteDB reloaded.')

                elif cmd == JM_CMD_SAVE_SITEDB:
                    globalObjects.siteDB.writeToFile()
            
                    self._send(JM_REPLY_OK)

                elif cmd == JM_CMD_GET_STATS_PROPERTIES:
                    if matcher is None:
                        self._send('')
                        NSLog(u'Mathcer is AWOL')
                    else:                    
                        self._send(matcher.tests.properties.getStatsString())

                elif cmd == JM_CMD_GET_STATS_PATTERNS:
                    if matcher is None:
                        self._send('')
                        NSLog(u'Mathcer is AWOL')
                    else:
                        self._send(matcher.tests.patterns.getStatsString())

                elif cmd == JM_CMD_SET_STATS_PROPERTIES:                    
                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        self._send(JM_REPLY_OK)
                        matcher.tests.properties.setStats(arg)

                elif cmd == JM_CMD_SET_STATS_PATTERNS:
                    if matcher is None:
                        self._send(JM_REPLY_ERR)
                        NSLog(u'Mathcer is AWOL')
                    else:
                        self._send(JM_REPLY_OK)            
                        matcher.tests.patterns.setStats(arg)
            
                else:
                    self._send(JM_REPLY_NA)

            except Exception, e:
                printException(u'Exception when JMServer is configuring', e)
            
            serverOpFinishedEvent.set()


class JMServer (SocketServer.ThreadingTCPServer):
    """JunkMatcher server collecting incoming messages and processing them"""
    def __init__ (self):
        """Attempts to find a port from PORT_MIN to PORT_MAX"""
        port = PORT_MIN
        while port < PORT_MAX:
            try:
                SocketServer.ThreadingTCPServer.__init__(self, ('127.0.0.1', port), JMHandler)
                break
            except:
                port += 1
        if port > PORT_MAX:
            open(TMP_PATH + 'server.port','w').write('ERR')
            NSLog(u'JMServer cannot find a good port between %d and %d.' % (PORT_MIN, PORT_MAX))
        else:
            self.request_queue_size = JMSERVER_QSIZE            
            open(TMP_PATH + 'server.port','w').write('%d' % port)
            NSLog(u'JunkMatcher Engine %s listening at port %d.' % (JM_ENGINE_VER, port))


# for safety
socket.setdefaulttimeout(JM_SOCK_TIMEOUT_VALUE)

jmServer = JMServer()
try:
    jmServer.serve_forever()
except Exception, e:
    try:
        os.remove(TMP_PATH + 'server.port')
    except:
        pass
    
    if isinstance(e, KeyboardInterrupt):
        NSLog(u'JMServer is interrupted')
    else:
        printException(u'JMServer made a boo-boo', e)
